/* Module      : JaegerDraw.cpp
 * Author      : Robert W. Lindeman
 * Email       : gogo@wpi.edu
 * Course      : IMGD 3000
 *
 * Description : This file contains the routines for drawing the scenes
 *               for the player.
 *
 * Date        : 2006/02/05
 *
 * History:
 * Revision      Date          Changed By
 * --------      ----------    ----------
 * 00.01         2006/02/05    gogo
 * First release.
 *
 * (c) Copyright 2006, Worcester Polytechnic Institute.
 */

/* -- INCLUDE FILES ------------------------------------------------------ */
 
#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
  #include<windows.h>
#endif

#ifdef MACOSX
  #include <gl.h>
  #include <glu.h>
  #include <glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif

#include "JaegerDraw.h"
#include "JaegerGlobals.h"
#include "JaegerConstants.h"
#include "BMPImg.h"

#include<iostream>
using namespace std;

/* -- DATA STRUCTURES ---------------------------------------------------- */

// For texturing.
typedef struct  {      // Create A Structure
  GLubyte *imageData;  // Image Data (Up To 32 Bits)
  GLuint   bpp;	       // Image Color Depth In Bits Per Pixel.
  GLuint   width;      // Image Width
  GLuint   height;     // Image Height
  GLuint   texID;      // Texture ID Used To Select A Texture
} TextureImage;        // Structure Name

// GUI elementes that can be displayed.
struct GUIStruct  {
  int where;
  const char *prompt;
  char str[256];
} GUIObjects[NUM_GUI_OBJECTS];

/* -- GLOBAL VARIABLES --------------------------------------------------- */
GLuint Celltex[3]; // 0 = wall, 1 = floor, 2 = ceiling

float HalfScreenWidth = ( ScreenWidth * 0.5 );

/* -- LOCAL VARIABLES ---------------------------------------------------- */
static void JaegerDrawMaze( void );
static void JaegerDrawCell( Cell *cell );
static void JaegerCube( float, float );
static void JaegerBuildCharacter( GLuint base );
static void JaegerDrawString( float x, float y, char *string );
static void JaegerDrawGUI( void );
extern void JaegerDrawFlashLight( void );
extern GLuint LoadGLTextures( char *file );
extern void JaegerDrawFlashlightObject();

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawAxes( void )
 *
 * Description : This function draws orthogonal axes in red, green, and
 *               blue at the current location.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerDrawAxes( void )  {
  GLUquadricObj *qobj;
  GLfloat cylRad = 0.025;
  GLfloat coneRad = 0.075;

  glPushMatrix( );
    // Do the X axis.
    glColor3f( 1.0, 0.0, 0.0 );
    glutSolidSphere( cylRad, 16, 16 );

    glRotated( 90.0, 0.0, 1.0, 0.0 );
    qobj = gluNewQuadric( );
    gluQuadricDrawStyle( qobj, GLU_FILL );
    gluCylinder( qobj, cylRad, cylRad, 1.0, 16, 1 );
    glTranslated( 0.0, 0.0, 1.0 );
    glutSolidCone( coneRad, 0.2, 16, 1 );
  glPopMatrix( );

  glPushMatrix( );
    // Do the Y axis.
    glColor3f( 0.0, 1.0, 0.0 );
    glRotated( -90.0, 1.0, 0.0, 0.0 );
    qobj = gluNewQuadric( );
    gluQuadricDrawStyle( qobj, GLU_FILL );
    gluCylinder( qobj, cylRad, cylRad, 1.0, 16, 1 );
    glTranslated( 0.0, 0.0, 1.0 );
    glutSolidCone( coneRad, 0.2, 16, 1 );
  glPopMatrix( );

  glPushMatrix( );
    // Do the Z axis.
    glColor3f( 0.0, 0.0, 1.0 );
    qobj = gluNewQuadric( );
    gluQuadricDrawStyle( qobj, GLU_FILL );
    gluCylinder( qobj, cylRad, cylRad, 1.0, 16, 1 );
    glTranslated( 0.0, 0.0, 1.0 );
    glutSolidCone( coneRad, 0.2, 16, 1 );
  glPopMatrix( );
}

void Cube(float ln)
{
  glBegin( GL_QUADS );
    // front face
    glNormal3f( 0.0, 0.0, 1.0 ); glTexCoord2f( 1.0, 1.0 ); glVertex3f( ln, ln, ln );
    glNormal3f( 0.0, 0.0, 1.0 ); glTexCoord2f( 0.0, 1.0 ); glVertex3f(-ln, ln, ln );
    glNormal3f( 0.0, 0.0, 1.0 ); glTexCoord2f( 0.0, 0.0 ); glVertex3f(-ln,-ln, ln );
    glNormal3f( 0.0, 0.0, 1.0 ); glTexCoord2f( 1.0, 0.0 ); glVertex3f( ln,-ln, ln );
    // back face
    glNormal3f( 0.0, 0.0,-1.0 ); glTexCoord2f( 1.0, 1.0 ); glVertex3f( ln, ln,-ln );
    glNormal3f( 0.0, 0.0,-1.0 ); glTexCoord2f( 0.0, 1.0 ); glVertex3f(-ln, ln,-ln );
    glNormal3f( 0.0, 0.0,-1.0 ); glTexCoord2f( 0.0, 0.0 ); glVertex3f(-ln,-ln,-ln );
    glNormal3f( 0.0, 0.0,-1.0 ); glTexCoord2f( 1.0, 0.0 ); glVertex3f( ln,-ln,-ln );
    // right face
    glNormal3f( 1.0, 0.0, 0.0 ); glTexCoord2f( 0.0, 1.0 ); glVertex3f( ln, ln, ln );
    glNormal3f( 1.0, 0.0, 0.0 ); glTexCoord2f( 0.0, 0.0 ); glVertex3f( ln,-ln, ln );
    glNormal3f( 1.0, 0.0, 0.0 ); glTexCoord2f( 1.0, 0.0 ); glVertex3f( ln,-ln,-ln );
    glNormal3f( 1.0, 0.0, 0.0 ); glTexCoord2f( 1.0, 1.0 ); glVertex3f( ln, ln,-ln );
    // left face
    glNormal3f( -1.0, 0.0, 0.0 ); glTexCoord2f( 0.0, 1.0 ); glVertex3f(-ln, ln, ln );
    glNormal3f( -1.0, 0.0, 0.0 ); glTexCoord2f( 0.0, 0.0 ); glVertex3f(-ln,-ln, ln );
    glNormal3f( -1.0, 0.0, 0.0 ); glTexCoord2f( 1.0, 0.0 ); glVertex3f(-ln,-ln,-ln );
    glNormal3f( -1.0, 0.0, 0.0 ); glTexCoord2f( 1.0, 1.0 ); glVertex3f(-ln, ln,-ln );
    // top face
    glNormal3f( 0.0, 1.0, 0.0 ); glTexCoord2f( 0.0, 1.0 ); glVertex3f( ln, ln, ln );
    glNormal3f( 0.0, 1.0, 0.0 ); glTexCoord2f( 1.0, 1.0 ); glVertex3f( ln, ln,-ln );
    glNormal3f( 0.0, 1.0, 0.0 ); glTexCoord2f( 1.0, 0.0 ); glVertex3f(-ln, ln,-ln );
    glNormal3f( 0.0, 1.0, 0.0 ); glTexCoord2f( 0.0, 0.0 ); glVertex3f(-ln, ln, ln );
    // bottom face
    glNormal3f( 0.0, -1.0, 0.0 ); glTexCoord2f( 0.0, 1.0 ); glVertex3f( ln,-ln, ln );
    glNormal3f( 0.0, -1.0, 0.0 ); glTexCoord2f( 1.0, 1.0 ); glVertex3f( ln,-ln,-ln );
    glNormal3f( 0.0, -1.0, 0.0 ); glTexCoord2f( 1.0, 0.0 ); glVertex3f(-ln,-ln,-ln );
    glNormal3f( 0.0, -1.0, 0.0 ); glTexCoord2f( 0.0, 0.0 ); glVertex3f(-ln,-ln, ln );
  glEnd( );
  
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawFlashLightObject( void )
 *
 * Description : This function draws the physical flashlight.
 *
 * Parameters  : void
 *
 * Returns     : void
 */
void JaegerDrawFlashlightObject( void )  {
  glMatrixMode( GL_MODELVIEW_MATRIX );
  glPushMatrix( );
    glLoadIdentity( );

    // put flashlight in front of the camera
    glTranslated( 0.15, -0.3, -0.4 );

    // disable lighting on the lightbulb
    glDisable( GL_LIGHTING );
    glColor3f( 1.0, 1.0, 1.0 ); // color it white
    glutSolidSphere( 0.06, 16, 16 );        
    glTranslated( 0.0, 0.0, 0.01 );
       
    // dark blue color
    glColor3f( 0.0, 0.0, 0.3 );
    GLUquadricObj* flashlighthandle = gluNewQuadric( ); // handle
    gluCylinder( flashlighthandle, 0.05, 0.05, 0.3, 16 , 1 );
    glutSolidCone( 0.07, 0.1, 16, 1 ); // container for the bulb
    glTranslated( 0.0, 0.045, 0.1);
    glColor3f( 0.0, 0.3, 0.5);
    glutSolidSphere( 0.015, 8, 8 ); // on/off button
    // enable lighting on the rest of the flashlight
    glEnable( GL_LIGHTING );
  glPopMatrix( );
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerCube( float length, float div )
 *
 * Description : This function draws a length float cube with texture
 *               coord specified.
 *               All cubes are length 'length'.
 * Parameters  : float length : Length of a cude side.
 *               float div    : Number of times the cube side will be
 *                              subdivided.
 *
 * Returns     : void
 */

void JaegerCube( float length , float div )  {
  float nlength = length / div;
  float ln = nlength * 0.5;
  float s = -length * 0.5;
  float e = -s;
  Cube( length * 0.5 );
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerBuildCharacter( GLuint base )
 *
 * Description : This function build up a display list of characters.
 *
 * Parameters  : GLuint : Base ??
 *
 * Returns     : void
 */
void JaegerBuildCharacter( GLuint base )  {
  for( int i = 0; i < 256; i++ )  {
    glNewList( base + i, GL_COMPILE );
    glutBitmapCharacter( FontType, i );
    glEndList( );
  }
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawString( float x, float y, char *string )
 *
 * Description : This function draws the string input.
 *
 * Parameters  : float x      : X location of string.
 *               float y      : Y location of string.
 *               char *string : String of chars to  print.
 *
 * Returns     : void
 */
void JaegerDrawString( float x, float y, char *string )  {
  glDisable( GL_LIGHTING );
  glMatrixMode( GL_MODELVIEW );
 
  glPushMatrix( );
    glLoadIdentity( );
    gluOrtho2D( 0, ScreenWidth, 0, ScreenHeight ); // for easy coordinate system
  
    glMatrixMode( GL_PROJECTION ); // set to PROJECTION mode
    glPushMatrix( );
      glLoadIdentity( );  
      glColor3f( 1.0, 1.0, 1.0 );
      glRasterPos2f( x, y );      
      glCallLists( ( GLint )strlen( string ), GL_UNSIGNED_BYTE, string );
  
    glPopMatrix( );
 
    glMatrixMode( GL_MODELVIEW ); // set to MODEL mode
  glPopMatrix( );
 
  glEnable( GL_LIGHTING );
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawGUI( void )
 *
 * Description : This function draws all the GUI elements.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerDrawGUI( void )  {
  char str[256];
  int strLen;
  int mid;

  // Where to start drawing.
  int topBase    = ScreenHeight - 20;
  int bottomBase = 10;
  int leftBase   = 5;
  int rightBase  = 0;

  // For each GUI element.
  for( int i = 0; i < NUM_GUI_OBJECTS; i++ )  {
    if( strlen( GUIObjects[ i ].str ) == 0 )  {
      continue;
    }
    sprintf( str, "%s%s", GUIObjects[ i ].prompt, GUIObjects[ i ].str );

    strLen = glutBitmapLength( FontType, ( const unsigned char * )str );
    mid = ( int )( HalfScreenWidth - ( strLen * 0.5 ) );

    rightBase  = ScreenWidth - ( strLen + leftBase );

    switch( GUIObjects[i].where )  {
      case GUI_UPPER_RIGHT:
        JaegerDrawString( rightBase, topBase, str );
        break;
      case GUI_LOWER_RIGHT:
        JaegerDrawString( rightBase, bottomBase, str );
        break;
      case GUI_UPPER_LEFT:
        JaegerDrawString( leftBase, topBase, str );
        break;
      case GUI_LOWER_LEFT:
        JaegerDrawString( leftBase, bottomBase, str );
        break;
      case GUI_CENTER_TOP:
        JaegerDrawString( mid, topBase, str );
        break;
      case GUI_CENTER_BOTTOM:
        JaegerDrawString( mid, bottomBase, str );
        break;
    }
  }
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerUIShow( int what, int where, char *str )
 *
 * Description : This function draw the string passed in where.
 *
 * Parameters  : int what  : What object is this?
 *               int where : Where do you want the string to go?
 *               char *str : String to draw.
 *
 * Returns     : void
 */

void JaegerUIShow( int what, int where, char *str )  {
  // Remember the string.
  strcpy( GUIObjects[what].str, str );
  // ...and where it goes.
  GUIObjects[what].where = where;
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawGoolie( int row, int col )
 *
 * Description : This function draws a Goolie in cell ( 'row', 'col' ).
 *
 * Parameters  : int row : Row to draw Goolie in.
 *               int col : Column to draw Goolie in.
 *
 * Returns     : void
 */

void JaegerDrawGoolie( int row, int col )  {
  printf( "Goolie is in cell: ( %d, %d ).\n", row, col );
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDraw( void )
 *
 * Description : This function draws the entire world.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerDraw( void )  {

// Hack, so that the camera is placed in the right place.
//  Cam.Slide( 0.0, 0.0, 0.0 );

  GLfloat matAmb[]   = { 0.4, 0.4, 0.4, 1.0 };
  GLfloat matDiff[]  = { 0.6, 0.6, 0.6, 1.0 };
  GLfloat matSpec[]  = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat matShine[] = { 50.0 };
  GLfloat lmAmb[]    = { 0.2, 0.2, 0.2, 1.0 };
  
  glMaterialfv( GL_FRONT, GL_AMBIENT, matAmb );
  glMaterialfv( GL_FRONT, GL_DIFFUSE, matDiff );
  glMaterialfv( GL_FRONT, GL_SPECULAR, matSpec );
  glMaterialfv( GL_FRONT, GL_SHININESS, matShine );
  
  GLfloat lInten[] = { 0.7, 0.7, 0.7, 1.0 };  
  
  glLightfv( GL_LIGHT1, GL_DIFFUSE, lInten );
  glLightfv( GL_LIGHT0, GL_DIFFUSE, lInten );
  glLightModelfv( GL_LIGHT_MODEL_AMBIENT, lmAmb );
  
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  glColor3f( 1.0, 1.0, 1.0 );
  if( DrawFlashLight )  {
    JaegerDrawFlashLight( );
    JaegerDrawFlashlightObject( );
  }  

  glEnable( GL_TEXTURE_2D );
  JaegerDrawMaze( );
  glDisable( GL_TEXTURE_2D );
  
  JaegerDrawGUI( );

  // Display the buffer we just rendered into.
  glutSwapBuffers( );
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawMaze( void )
 *
 * Description : This function draws the labyrinth.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

static void JaegerDrawMaze( void )  {
  for( int row = 0; row < GertMaze->height; row++ )  {
    for( int col = 0; col < GertMaze->width; col++ )  {
      glPushMatrix( );
        glTranslated( col * CellSide,
                      0.0,
                      row * CellSide );
        JaegerDrawCell( GertMaze->GetCell( row, col ) );
      glPopMatrix( );
    }
  }
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawCell( Cell *cell )
 *
 * Description : This function draws the cell associated with 'cell'.
 *      ________
 *     /       /|
 *    /       / |
 *   /_______/  |
 *   |   /   |  |
 *   |  /    |  /
 *   | /     | /
 *   |/______|/
 * 
 *  There are 15 different cell types (not including the door type).
 *  However, these can be reduced to 5 distinct cell types, with the
 *  others being simple rotations of one of these five. Therefore,
 *  find out which "class" of cell we need to draw, apply any necessary
 *  rotation, and then draw that class of cell.
 *
 * Parameters  : Cell cell : Cell to draw.
 *
 * Returns     : void
 */

static void JaegerDrawCell( Cell *cell )  {

  /*
  if( cell == CurCell )  {
    // Draw axes.
    glPushMatrix( );
      glScaled( 75.0, 75.0, 75.0 );
      JaegerDrawAxes( );
    glPopMatrix( );
  }
  */
  float subdivide = 10;
  float polecolor[] = { 0.1, 0.1, 0.1 };
  switch( cell->type )  {
    case CELL_CROSS: // This is its own class.
      // Do the pole at the North-West corner.
      glPushMatrix( );
        glDisable( GL_TEXTURE_2D );
        glColor3fv( polecolor );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the pole at the North-East corner.
      glPushMatrix( );
        glColor3fv( polecolor );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the pole at the South-West corner.
      glPushMatrix( );
        glColor3fv( polecolor );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the pole at the South-East corner.
      glPushMatrix( );
        glColor3fv(polecolor);
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the floor.
      glBindTexture( GL_TEXTURE_2D, Celltex[1] );
      glEnable( GL_TEXTURE_2D );

      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      break;

    case CELL_VERT:
    case CELL_HORIZ:
      // Now apply the correct rotation.
      switch( cell->type )  {
        case CELL_VERT:
          break;
        case CELL_HORIZ:
          glRotated( 90.0, 0.0, 1.0, 0.0 );
          break;
      }
      // Do the West wall.
      glBindTexture( GL_TEXTURE_2D, Celltex[0] );

      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the East wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the floor.
      glBindTexture( GL_TEXTURE_2D, Celltex[1] );

      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      break;

    case CELL_SOUTH_WEST:
    case CELL_SOUTH_EAST:
    case CELL_NORTH_WEST:
    case CELL_NORTH_EAST:
      // Now apply the correct rotation.
      switch( cell->type )  {
        case CELL_SOUTH_WEST:
          break;
        case CELL_SOUTH_EAST:
          glRotated( 90.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_NORTH_WEST:
          glRotated( -90.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_NORTH_EAST:
          glRotated( 180.0, 0.0, 1.0, 0.0 );
          break;
      }
// Adjust the wall sizes for the overlap at corners.
      // Do the West wall.
      glBindTexture( GL_TEXTURE_2D, Celltex[0] );

      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the South wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  0.0,
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( CellSide, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the pole at the North-East corner.
	  glDisable(GL_TEXTURE_2D);
      glPushMatrix( );
        glColor3fv(polecolor);
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the floor.
      glEnable( GL_TEXTURE_2D );
      glBindTexture( GL_TEXTURE_2D, Celltex[1] );

      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        JaegerCube( 1.0 , subdivide );
      glPopMatrix( );
      break;

    case CELL_NORTH_TEE:
    case CELL_EAST_TEE:
    case CELL_SOUTH_TEE:
    case CELL_WEST_TEE:
      // Now apply the correct rotation.
      switch( cell->type )  {
        case CELL_NORTH_TEE:
          break;
        case CELL_EAST_TEE:
          glRotated( -90.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_SOUTH_TEE:
          glRotated( 180.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_WEST_TEE:
          glRotated( 90.0, 0.0, 1.0, 0.0 );
          break;
      }
      // Do the North wall.
      glBindTexture( GL_TEXTURE_2D, Celltex[0] );

      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  0.0,
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( CellSide, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      glDisable( GL_TEXTURE_2D );
      // Do the pole at the South-West corner.
      glPushMatrix( );
        glColor3fv( polecolor );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the pole at the South-East corner.
      glPushMatrix( );
        glColor3fv( polecolor );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( WallThickness, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the floor.
      glEnable( GL_TEXTURE_2D );
      glBindTexture( GL_TEXTURE_2D, Celltex[1] );

      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        JaegerCube( 1.0 , subdivide );
      glPopMatrix( );
      break;

    case CELL_NORTH_DEADEND:
    case CELL_EAST_DEADEND:
    case CELL_SOUTH_DEADEND:
    case CELL_WEST_DEADEND:
      // Now apply the correct rotation.
      switch( cell->type )  {
        case CELL_NORTH_DEADEND:
          break;
        case CELL_EAST_DEADEND:
          glRotated( -90.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_SOUTH_DEADEND:
          glRotated( 180.0, 0.0, 1.0, 0.0 );
          break;
        case CELL_WEST_DEADEND:
          glRotated( 90.0, 0.0, 1.0, 0.0 );
          break;
      }
// Adjust the wall sizes for the overlap at corners.
      // Do the North wall.
      glBindTexture( GL_TEXTURE_2D, Celltex[0] );

      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  0.0,
                       ( CellHeight * 0.5 ),
                      -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ) );
        glScaled( CellSide, CellHeight, WallThickness );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the West wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated( -( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the East wall.
      glPushMatrix( );
        glColor3f( 1.0, 0.6, 0.2 );
        glTranslated(  ( ( CellSide * 0.5 ) - ( WallThickness * 0.5 ) ),
                       ( CellHeight * 0.5 ),
                       0.0 );
        glScaled( WallThickness, CellHeight, CellSide );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      // Do the floor.
      glBindTexture( GL_TEXTURE_2D, Celltex[1] );

      glPushMatrix( );
        glColor3f( 1.0, 0.7, 0.7 );
        glTranslated( 0.0, -WallThickness * 0.5, 0.0 );
        glScaled( CellSide, WallThickness, CellSide );
        JaegerCube( 1.0, subdivide );
      glPopMatrix( );
      break;

    default:
      break;
  }
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawInit( int argc, char *argv[] )
 *
 * Description : This function initializes the graphics state.
 *
 * Parameters  : int argc    : Number of command-line arguments.
 *               char *argv[]: Array of command-line arguments.
 *
 * Returns     : void
 */

void JaegerDrawInit( int argc, char *argv[] )  {
  glutInit( &argc, argv );
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  glutInitWindowSize( ScreenWidth, ScreenHeight );
  glutInitWindowPosition( 100, 100 );
  glutCreateWindow( WindowTitle );

  glutDisplayFunc( JaegerDraw );

  glShadeModel( GL_SMOOTH );
  glEnable( GL_DEPTH_TEST );
  glEnable( GL_NORMALIZE );
  glEnable( GL_BLEND );

  glEnable( GL_LIGHTING );
//  glEnable( GL_LIGHT0 );
  glEnable( GL_LIGHT1 );

  glColorMaterial( GL_FRONT, GL_DIFFUSE );
  glEnable( GL_COLOR_MATERIAL );
  glClearColor( 0.1, 0.1, 0.1, 0.0 );

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity( );
  Cam.SetShape( ViewAngle, AspectRatio, NearClip, FarClip );
  Cam.Set( EyePt, InterestPt, UpVec );
  
  // load texture
  Celltex[0] = LoadGLTextures( Walltex );
  Celltex[1] = LoadGLTextures( Floortex );
  Celltex[2] = LoadGLTextures( Ceilingtex );

  // Initialize GUI stuff.
  for( int i = 0; i < NUM_GUI_OBJECTS; i++ )  {
    GUIObjects[ i ].where = 0;
    GUIObjects[ i ].prompt = GUIPrompts[i];
    strcpy( GUIObjects[ i ].str, "" );
  }

  GLuint charBase = 0; // start of character list
  glGenLists( 256 );
  JaegerBuildCharacter( charBase );

  //glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
  //glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  //glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerTurnLeft90( void )
 *
 * Description : This function turns the users view 90 deg. left.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerTurnLeft90( void )  {
  for( int i = 0; i < JaegerMovement; i++ )  {
    Cam.Yaw( 90 / JaegerMovement );
    JaegerDraw( );
  }

  switch( Hedi->ori )  {
    case ORI_NORTH:
      Hedi->ori = ORI_WEST;
      break;
    case ORI_WEST:
      Hedi->ori = ORI_SOUTH;
      break;
    case ORI_SOUTH:
      Hedi->ori = ORI_EAST;
      break;
    case ORI_EAST:
      Hedi->ori = ORI_NORTH;
      break;
    default:
      break;
  }
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerTurnRight90( void )
 *
 * Description : This function turns the users view 90 deg. right.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerTurnRight90( void )  {
  for( int i = 0; i < JaegerMovement; i++ )  {
    Cam.Yaw( -90.0 / JaegerMovement );
    JaegerDraw( );
  }

  switch( Hedi->ori )  {
    case ORI_NORTH:
      Hedi->ori = ORI_EAST;
      break;
    case ORI_WEST:
      Hedi->ori = ORI_NORTH;
      break;
    case ORI_SOUTH:
      Hedi->ori = ORI_WEST;
      break;
    case ORI_EAST:
      Hedi->ori = ORI_SOUTH;
      break;
    default:
      break;
  }
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerMoveForwardCell( void )
 *
 * Description : This function moves the user 1 cell forward, if
 *               possible.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerMoveForwardCell( void )  {
  for( int i = 0; i < JaegerMovement; i++ )  {
    Cam.Slide( 0.0, 0.0, -CellSide / JaegerMovement );
    JaegerDraw( );
  }

  switch( Hedi->ori )  {
    case ORI_NORTH:
      Hedi->pos[Yidx]--;
      break;
    case ORI_WEST:
      Hedi->pos[Xidx]--;
      break;
    case ORI_SOUTH:
      Hedi->pos[Yidx]++;
      break;
    case ORI_EAST:
      Hedi->pos[Xidx]++;
      break;
    default:
      break;
  }
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerMoveBackCell( void )
 *
 * Description : This function moves the user 1 cell back, if
 *               possible.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerMoveBackCell( void )  {
  for( int i = 0; i < JaegerMovement; i++ )  {
    Cam.Slide( 0.0, 0.0, CellSide / JaegerMovement );
    JaegerDraw( );
  }

  switch( Hedi->ori )  {
    case ORI_NORTH:
      Hedi->pos[Yidx]++;
      break;
    case ORI_WEST:
      Hedi->pos[Xidx]++;
      break;
    case ORI_SOUTH:
      Hedi->pos[Yidx]--;
      break;
    case ORI_EAST:
      Hedi->pos[Xidx]--;
      break;
    default:
      break;
  }
}
  
/* ----------------------------------------------------------------------- */
/* Function    : void JaegerMainLoop( void )
 *
 * Description : This function enters the event loop for glut.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

int JaegerMainLoop( void )  {
  glutMainLoop( );
  return( 0 );
}

/* ----------------------------------------------------------------------- */
/* Function    : void JaegerDrawFlashLight( void )
 *
 * Description : This function draws flashlight with camera as light location.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void JaegerDrawFlashLight( void )  {
  glMatrixMode( GL_MODELVIEW_MATRIX );
  glPushMatrix( );
  glLoadIdentity( );
	
  glTranslated( 0.15, -0.3, -0.4);
  float lightpos[] = { 0.0, 0.0, 0.0, 1.0 };
  glLightfv( GL_LIGHT1, GL_POSITION, lightpos );
  
  // light direction
  float dir[] = { Lightdir.x, Lightdir.y, Lightdir.z };
  
  // set spot light parameters
  glLightfv( GL_LIGHT1, GL_SPOT_DIRECTION, dir );
  // set cutoff angle
  glLightf( GL_LIGHT1, GL_SPOT_CUTOFF, SpotCutoff );
  // set how fast the light drop off
  glLightf( GL_LIGHT1, GL_SPOT_EXPONENT, SpotExp < 0.0 ? SpotExp : 0.0 );

  // set up attenutation 1/(a+bd+cd^2)
  glLightf( GL_LIGHT1, GL_CONSTANT_ATTENUATION, SpotConstantAtt );  // a
  glLightf( GL_LIGHT1, GL_LINEAR_ATTENUATION, SpotLinearAtt );      // b
  glLightf( GL_LIGHT1, GL_QUADRATIC_ATTENUATION, SpotQuadAtt );     // c    

  glPopMatrix( );
}

/* ----------------------------------------------------------------------- */
/* Function    : int ImageLoad(char *filename, TextureImage *textture )
 *
 * Description : This function loads bitmap with 24 bit resolusion, it won't 
 *               work with non 24bit bitmap file!! code from Jeff Molofee '99
 *               Nehe's linux code for texture loading
 *
 * Parameters  : char *filename : Name of texture file.
 *               TextureImage *texture : Texture structure to fill.
 *
 * Returns     : int : 1 if success, all else 0.
 */

int ImageLoad( char *filename, TextureImage *texture )  {
  // Uncompressed TGA Header
  GLubyte TGAheader[12]={ 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  GLubyte TGAcompare[12]; // Used To Compare TGA Header
  GLubyte header[6];      // First 6 Useful Bytes From The Header
  GLuint bytesPerPixel;   // Holds Number Of Bytes Per Pixel Used In The TGA File
  GLuint imageSize;       // Used To Store The Image Size When Setting Aside Ram
  GLuint temp;            // Temporary Variable

// Open The TGA File
  FILE *file = fopen( filename, "rb" );

  // If the file does not exist, or there are fewer than 12 bytes to read, or the header is
  // wrong, or the header is too short, then fail.
  // If So Read Next 6 Header Bytes
  if( ( file == NULL ) ||
      ( fread( TGAcompare, 1, sizeof( TGAcompare ), file ) != sizeof( TGAcompare ) ) ||
      ( memcmp( TGAheader, TGAcompare, sizeof( TGAheader ) ) != 0 ) ||
      ( fread( header, 1, sizeof( header ), file ) != sizeof( header ) ) )  {
    fclose( file );   // If Anything Failed, Close The File
    return( false );  // Return False
  }
  
  texture->width  = header[1] * 256 + header[0]; // Determine The TGA Width  (highbyte*256+lowbyte)
  texture->height = header[3] * 256 + header[2]; // Determine The TGA Height (highbyte*256+lowbyte)
 
  if( ( texture->width <= 0 ) ||  // Is The Width Less Than Or Equal To Zero
      ( texture->height	<= 0 ) || // Is The Height Less Than Or Equal To Zero
      ( header[4] != 24 && header[4] != 32 ) )  {  // Is The TGA 24 or 32 Bit?
    fclose( file );               // If Anything Failed, Close The File
    return( false );              // Return False
  }
  
  texture->bpp = header[4];  // Grab The TGA's Bits Per Pixel (24 or 32)
  bytesPerPixel	= texture->bpp / 8; // Divide By 8 To Get The Bytes Per Pixel
  // Calculate The Memory Required For The TGA Data
  imageSize = texture->width * texture->height * bytesPerPixel;
  
  texture->imageData=( GLubyte * )malloc( imageSize );		// Reserve Memory To Hold The TGA Data
  
  // Does The Image Size Match The Memory Reserved?
  if( ( texture->imageData == NULL ) || // Does The Storage Memory Exist?
      ( fread( texture->imageData, 1, imageSize, file ) != imageSize ) )  {
    if( texture->imageData != NULL )  { // Was Image Data Loaded
      free( texture->imageData );       // If So, Release The Image Data
    }
  
    fclose( file );   // Close The File
    return( false );  // Return False
  }
  
  // Loop Through The Image Data
  for( GLuint i = 0; i < int( imageSize ); i += bytesPerPixel )  {
    // Temporarily Store The Value At Image Data 'i'
    temp=texture->imageData[i];
    // Set The 1st Byte To The Value Of The 3rd Byte
    texture->imageData[i] = texture->imageData[i + 2];
    // Set The 3rd Byte To The Value In 'temp' (1st Byte Value)
    texture->imageData[i + 2] = temp;
  }
  
  fclose( file );
  return( true );
}

/* ----------------------------------------------------------------------- */
/* Function    : GLuint LoadGLTextures( char *file )
 *
 * Description : This function loads bitmap file using loadBMP function and
 *               load it into openGL. copyed from Nehe
 *
 * Parameters  : char *file : Name of texture file to load.
 *
 * Returns     : GLuint : texture id
 */

GLuint LoadGLTextures( char *file )  {
  // Load Texture
  string f = file;
  bool useBMP, useTGA;
  BMPImg loader;
  TextureImage *image1;
  GLuint type = GL_RGBA; // Set The Default GL Mode To RBGA (32 BPP)
  useBMP = useTGA = false;

  if( f.find( ".bmp", 0 ) != string::npos )  {    
    if( loader.Load( file ) != 0x1 )  {
      printf( "%s load not successful\n" );
      exit( 0 );
    }
    useBMP = true;
  }
  else  {
    if( f.find( ".tga", 0 ) != string::npos )  {
      // allocate space for texture
      image1 = ( TextureImage * )malloc( sizeof( TextureImage ) );
      if( image1 == NULL )  {
	printf( "Error allocating space for image\n" );
	exit( 0 );
      }

      if( !ImageLoad( file, image1 ) )  {
	printf( "%s load failed\n", file );
	exit( 0 );
      }      
      useTGA = true;
      if( image1->bpp == 24 )  {  // Was The TGA 24 Bits
	type = GL_RGB;            // If So Set The 'type' To GL_RGB
      }
    }
    else  {
      printf( "%s is not a supported file format.\n", file );
      exit( 0 );
    }
  }
  GLuint texture[1];
  // Create Texture	
  glGenTextures( 1, &texture[0] );
  // 2d texture (x and y size)
  glBindTexture( GL_TEXTURE_2D, texture[0] );

  // scale linearly when image bigger than texture
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  // scale linearly when image smalled than texture
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

  // 2d texture, level of detail 0 (normal), 3 components (red, green, blue), x size from image, y size from image, 
  // border 0 (normal), rgb color data, unsigned byte data, and finally the data itself.
  //glTexImage2D(GL_TEXTURE_2D, 0, 3, image1->sizeX, image1->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, image1->data);
  if( useBMP )  {
    glTexImage2D( GL_TEXTURE_2D, 0, 3, loader.GetWidth(),
                  loader.GetHeight(), 0, GL_RGB, GL_UNSIGNED_BYTE, loader.GetImg( ) );
  }
  else  {
    if( useTGA )  {
      glTexImage2D( GL_TEXTURE_2D, 0, 3, image1->width, image1->height,
                    0, type, GL_UNSIGNED_BYTE, image1->imageData );
    }
  }

  return( texture[0] );
}

/* ----------------------------------------------------------------------- */
