/* Module      : JaegerMaze.cpp
 * Author      : Robert W. Lindeman
 * Email       : gogo@wpi.edu
 * Course      : IMGD 3000
 *
 * Description : This is the code for the Maze class.
 *
 *               A maze is made up of a two-dimentional array of 'Cell'
 *               objects.
 *
 * Date        : 2006/02/06
 *
 * History:
 * Revision      Date          Changed By
 * --------      ----------    ----------
 * 00.01         2006/02/06    gogo
 * First release.
 *
 * (c) Copyright 2006, Worcester Polytechnic Institute.
 */

/* -- INCLUDE FILES ------------------------------------------------------ */

#include <stdio.h>
#include "JaegerMaze.h"
#include <stdlib.h>

/* -- DATA STRUCTURES ---------------------------------------------------- */

/* -- GLOBAL VARIABLES --------------------------------------------------- */

/* -- LOCAL VARIABLES ---------------------------------------------------- */

/* -----------------------------------------------------------------------
 * Function    : Maze::Maze( int w, int h )
 *
 * Description : This function allocates space for a maze of width 'w'
 *               and height 'h'.
 *
 * Parameters  : int   w: Width of the maze to generate.
 *               int   h: Height of the maze to generate.
 *
 * Returns     : void
 */

Maze::Maze( int w, int h )  {
  this->width = w;
  this->height = h;

  // Allocate the first dimension.
  this->maze = ( Cell *** )malloc( h * sizeof( Cell * ) );

  if( this->maze )  {
    // Then add the cells.
    for( int r = 0; r < this->height; r++ )  {
      this->maze[r] = ( Cell ** )malloc( w * sizeof( Cell * ) );
      for( int c = 0; c < this->width; c++ )  {
        this->maze[r][c] = new Cell;
      }
    }
  }
}

/* -----------------------------------------------------------------------
 * Function    : Maze::Maze( void )
 *
 * Description : This function allcates space for a maze based on data
 *               read from a file, and then loads the data into the maze.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

Maze::Maze( void )  {
  this->width  = 0;
  this->height = 0;
  this->maze   = NULL;
}

/* -----------------------------------------------------------------------
 * Function    : Maze::~Maze( void )
 *
 * Description : This function frees the existing maze.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

Maze::~Maze( void )  {
  if( this->maze )  {
    for( int r = 0; r < this->height; r++ )  {
      for( int c = 0; c < this->width; c++ )  {
        delete this->maze[r][c];
      }
      free( this->maze[r] );
    }

    free( this->maze );
  }
}

/* -----------------------------------------------------------------------
 * Function    : bool Maze::GenerateDefaultMaze( float s )
 *
 * Description : This function builds a default maze layout.
 *
 * Parameters  : float s : Random-number seed (Not used here).
 *
 * Returns     : bool : true = Success, false = Error.
 */

bool Maze::GenerateDefaultMaze( float s )  {
  this->seed = s;

  // For now, just create one that has the following shape:
  //   _____ _____
  //  |_____|_____|
  //  |_____|_____|

  // The width and height need to be greater than 1 for this to work.
  if( ( this->width < 2 ) || ( this->height < 2 ) )  {
    return( false );
  }

  // The corners are easy.
  this->maze[0][0]->type = CELL_NORTH_WEST;
  this->maze[0][this->width-1]->type = CELL_NORTH_EAST;
  this->maze[this->height-1][this->width-1]->type = CELL_SOUTH_EAST;
  this->maze[this->height-1][0]->type = CELL_SOUTH_WEST;

  // First, find the "middle" of each dimension.
  int midCol = ( int )( this->width  * 0.5 );
  int midRow = ( int )( this->height * 0.5 );

  // Now do the first and last rows.
  for( int c = 1; c < this->width-1; c++ )  {
    // Add the Tee-Junctions, if necessary.
    if( c == midCol )  {
      this->maze[0][c]->type = CELL_NORTH_TEE;
      this->maze[this->height-1][c]->type = CELL_SOUTH_TEE;
      // Fill in the middle column with verticals.
      for( int r = 1; r < this->height-1; r++ )  {
        this->maze[r][c]->type = CELL_VERT;
      }
    }
    else  {
      // Otherwise, just make them horizontals.
      this->maze[0][c]->type = CELL_HORIZ;
      this->maze[this->height-1][c]->type = CELL_HORIZ;
    }
  }

  // Now do the first and last columns.
  for( int r = 1; r < this->height-1; r++ )  {
    // Add the Tee-Junctions, if necessary.
    if( r == midRow )  {
      this->maze[r][0]->type = CELL_WEST_TEE;
      this->maze[r][this->width-1]->type = CELL_EAST_TEE;
      // Fill in the middle row with horizontals.
      for( int c = 1; c < this->width-1; c++ )  {
        this->maze[r][c]->type = CELL_HORIZ;
      }
    }
    else  {
      // Otherwise, just make them verticals.
      this->maze[r][0]->type = CELL_VERT;
      this->maze[r][this->width-1]->type = CELL_VERT;
    }
  }

  // If both dimensions are greater than 2, add the "cross" in the middle.
  if( ( this->width > 2 ) && ( this->height > 2 ) )  {
    this->maze[midRow][midCol]->type = CELL_CROSS;
  }

  return( true );
}

/* -----------------------------------------------------------------------
 * Function    : bool Maze::GenerateRandomMaze( float s )
 *
 * Description : This function builds a random maze layout based on the
 *               random number seed 's'.
 *
 * Parameters  : float s : Random-number seed to use for choose cell types.
 *
 * Returns     : bool : true = Success, false = Error.
 */

bool Maze::GenerateRandomMaze( float s )  {
  this->seed = s;

  // This is where the code goes to generate your maze layout, based on
  // the random number seed.

  return( true );
}

/* -----------------------------------------------------------------------
 * Function    : void Maze::Print( void )
 *
 * Description : This function outputs the current maze to stdout.
 *
 * Parameters  : void
 *
 * Returns     : void
 */

void Maze::Print( void )  {
  char ch;
/*
  char extAscii[15] =  {
    0xCE, // CELL_CROSS
    0xBA, // CELL_VERT
    0xCD, // CELL_HORIZ
    0xC8, // CELL_SOUTH_WEST
    0xBC, // CELL_SOUTH_EAST
    0xC9, // CELL_NORTH_WEST
    0xBB, // CELL_NORTH_EAST
    0xCB, // CELL_NORTH_TEE
    0xB9, // CELL_EAST_TEE
    0xCA, // CELL_SOUTH_TEE
    0xCC, // CELL_WEST_TEE
    0xD2, // CELL_NORTH_DEADEND
    0xB5, // CELL_EAST_DEADEND
    0xD0, // CELL_SOUTH_DEADEND
    0xC6  // CELL_WEST_DEADEND
  };
*/
  char extAscii[15] =  {
    '+',  // CELL_CROSS
    '|',  // CELL_VERT
    '-',  // CELL_HORIZ
    '\\', // CELL_SOUTH_WEST
    '/',  // CELL_SOUTH_EAST
    '/',  // CELL_NORTH_WEST
    '\\', // CELL_NORTH_EAST
    '+',  // CELL_NORTH_TEE
    '+',  // CELL_EAST_TEE
    '+',  // CELL_SOUTH_TEE
    '+',  // CELL_WEST_TEE
    '.',  // CELL_NORTH_DEADEND
    '.',  // CELL_EAST_DEADEND
    '.',  // CELL_SOUTH_DEADEND
    '.'   // CELL_WEST_DEADEND
  };

  if( !this->maze )  {
    return;
  }

  // Now print it row by row.
  for( int r = 0; r < this->height; r++ )  {
    for( int c = 0; c < this->width; c++ )  {
      if( this->maze[r][c]->type == CELL_NONE )  {
        ch = ' '; // 0x20;
      }
      else  {
        ch = extAscii[this->maze[r][c]->type];
      }
      fprintf( stdout, "%c", ch );
    }
    fprintf( stdout, "\n" );
  }
}

/* -----------------------------------------------------------------------
 * Function    : int Maze::Load( char *file )
 *
 * Description : This function allcates space for a maze based on data
 *               read from a file, and then loads the data into the maze.
 *
 * Parameters  : char *file : File to load maze from.
 *
 * Returns     : int : 0 = Success. -1 = Error.
 */

int Maze::Load( char *file )  {
  // To avoid memory leaks, remove any old cells.
  if( this->maze )  {
    for( int r = 0; r < this->height; r++ )  {
      for( int c = 0; c < this->width; c++ )  {
        delete this->maze[r][c];
      }
      free( this->maze[r] );
    }
    free( this->maze );
  }
  this->width  = 0;
  this->height = 0;
  this->maze   = NULL;

  // Place your code here to open the file, and read in the data to
  // allocate the maze, and to load the saved maze layout into.

  return( 0 );
}

/* -----------------------------------------------------------------------
 * Function    : int Maze::Save( char *file )
 *
 * Description : This function saves the current maze into a file.
 *
 * Parameters  : char *file : File to save maze to.
 *
 * Returns     : int : 0 = Success. -1 = Error.
 */

int Maze::Save( char *file )  {
  // Place your code here to open the file, and save the data to.
  return( 0 );
}

/* ----------------------------------------------------------------------- */

